package org.jvnet.transcoder.maven2;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Pattern;

import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.jvnet.transcoder.maven2.util.PropertyUtils;

/**
 * Verifies the completeness of multilingual resources and generates property
 * files with missing entries. Accepts locales and resource names (<code>com/acme/foo/messages</code>,
 * without trailing <code>_<em>locale</em>.properties</code>) as
 * configuration. Checks all the runtime dependencies of the project. Generates
 * property files with missing entries in
 * <code><em>groupId</em>/<em>artifactId</em></code> inside the target
 * directory.
 * 
 * @goal verify
 * @phase verify
 */
public class TranscoderMojo extends AbstractMojo {

	/**
	 * The Maven project reference, used internally.
	 * 
	 * @parameter expression="${project}"
	 * @required
	 * @readonly
	 */
	protected MavenProject project;

	/**
	 * Used internally to resolve {@link #plugins} to their jar files.
	 * 
	 * @component
	 */
	protected ArtifactResolver artifactResolver;

	/**
	 * Used internally to produce dependency artifacts.
	 * 
	 * @component
	 */
	protected ArtifactFactory artifactFactory;

	/**
	 * Used internally for artifact resolution.
	 * 
	 * @parameter expression="${localRepository}"
	 * @required
	 */
	protected ArtifactRepository localRepository;

	/**
	 * Locale codes.
	 * 
	 * @parameter
	 * @required
	 */
	protected String[] locales = new String[] {};

	/**
	 * Resource names.
	 * 
	 * @parameter
	 * @required
	 */
	protected String[] resources = new String[] {};

	/**
	 * Target directory.
	 * 
	 * @parameter default-value="${project.build.directory}/missing-resources"
	 * @required
	 */
	protected File target;

	/**
	 * Regular expression for "todo" entries. If entry value matches this
	 * expression, it will be considered missing.
	 * 
	 * @parameter
	 */
	protected String missingResourceRegex;

	public void execute() throws MojoExecutionException, MojoFailureException {
		final List runtimeDependencies = project.getRuntimeDependencies();

		final Locale[] locales = parseLocales(this.locales);

		for (Iterator iterator = runtimeDependencies.iterator(); iterator
				.hasNext();) {
			final Dependency dependency = (Dependency) iterator.next();

			final org.apache.maven.artifact.Artifact artifact = artifactFactory
					.createArtifact(dependency.getGroupId(), dependency
							.getArtifactId(), dependency.getVersion(),
							dependency.getScope(), dependency.getType());

			try {
				artifactResolver.resolve(artifact, project
						.getRemoteArtifactRepositories(), localRepository);
				final URL url = artifact.getFile().toURL();
				for (int jndex = 0; jndex < resources.length; jndex++) {
					final String resource = resources[jndex];
					processDependencyResources(dependency, url, resource,
							locales);
				}

			} catch (Exception ex) {

			}

		}
	}

	public void processDependencyResources(final Dependency dependency,
			URL url, String name, Locale[] locales) {
		final Map[] maps = new Map[locales.length];
		final Set codes = new TreeSet();
		for (int index = 0; index < locales.length; index++) {
			final Locale locale = locales[index];
			maps[index] = loadResourceProperties(url, name, locale);
			codes.addAll(maps[index].keySet());
		}

		final Map[] missingMaps = new Map[locales.length];
		for (int index = 0; index < locales.length; index++) {
			final Locale locale = locales[index];
			final Map map = maps[index];
			missingMaps[index] = new TreeMap();
			for (final Iterator iterator = codes.iterator(); iterator.hasNext();) {
				final String code = (String) iterator.next();
				final String value = (String) map.get(code);
				if (isMissingProperty(code, value)) {
					missingMaps[index].put(code, getDefaultValue(code, value));
				}
			}
			if (!missingMaps[index].isEmpty()) {
				try {
					final File file = new File(target, dependency.getGroupId()
							+ File.separator + dependency.getArtifactId()
							+ File.separator + getResourceName(name, locale));
					file.getParentFile().mkdirs();
					final FileOutputStream fos = new FileOutputStream(file);
					PropertyUtils.store(missingMaps[index], fos, "");
					fos.close();
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			}
		}
	}

	private Pattern missingPropertyPattern;

	protected boolean isMissingProperty(final String code, final String value) {
		if (missingPropertyPattern == null && missingResourceRegex != null) {
			missingPropertyPattern = Pattern.compile(missingResourceRegex);
		}
		return value == null
				|| (missingPropertyPattern != null && missingPropertyPattern
						.matcher(value).matches());
	}

	public Map loadResourceProperties(URL url, String name, Locale locale) {
		final ClassLoader cl = new URLClassLoader(new URL[] { url });

		final String resourceName = getResourceName(name, locale);
		final URL resourceURL = cl.getResource(resourceName);

		if (resourceURL == null) {
			return Collections.EMPTY_MAP;
		} else {
			final InputStream resourceStream = cl
					.getResourceAsStream(resourceName);
			try {
				return PropertyUtils.load(resourceStream);
			} catch (IOException ex) {
				getLog()
						.warn(
								"Could not load properties from ["
										+ resourceURL + "].", ex);
				return Collections.EMPTY_MAP;
			} finally {
				try {
					resourceStream.close();
				} catch (IOException ignored) {

				}

			}
		}
	}

	protected String getDefaultValue(final String code, final String value) {
		return value == null ? "@@TODO." + code + "@@" : value;
	}

	protected String getResourceName(String name, Locale locale) {
		return name
				+ ((locale == null || "".equals(locale)) ? "" : "_"
						+ locale.toString()) + ".properties";
	}

	private Locale[] parseLocales(String[] localeStrings) {
		final Locale[] locales = new Locale[localeStrings.length];
		for (int index = 0; index < localeStrings.length; index++) {
			locales[index] = parseLocale(localeStrings[index]);
		}
		return locales;
	}

	private Locale parseLocale(String string) {
		if (string == null || "".equals(string)) {
			return null;
		} else {
			final StringTokenizer tokens = new StringTokenizer(string, "_");
			final String language = tokens.hasMoreTokens() ? tokens.nextToken()
					: "";
			final String country = tokens.hasMoreTokens() ? tokens.nextToken()
					: "";
			String variant = "";
			String sep = "";
			while (tokens.hasMoreTokens()) {
				variant += sep + tokens.nextToken();
				sep = "_";
			}
			return new Locale(language, country, variant);
		}
	}

}
